
# Example: Publish a Version to Github

Publishing a version tag for your project in Github requires a few extra steps.
`devops` can automate that for you by using information from your `package.json` file.

### my-tasks.js

The task file is a Javascript file which defines your tasks as data constructs.
This file is used by `devops` to run your tasks.

```js
module.exports = {

	Context:{
		// Load the project's package file into the Context.
		Package: require( 'package.json' ),
	},

	git_publish_version: [
		{ // Stash all changes.
			$Shell: {
				command: 'git add .',
				out: { console: true },
				err: { console: true },
				halt_on_error: false
			}
		},
		{ // Commit all changes with a custom message.
			$Shell: {
				command: 'git commit --quiet -m "Finalization for v${Package.version}"',
				out: { console: true },
				err: { console: true },
				halt_on_error: false
			}
		},
		{ // Push all changes to Github.
			$Shell: {
				command: 'git push --quiet origin main',
				out: { console: true },
				err: { console: true },
				halt_on_error: false
			}
		},
		{ // Create a tag for this version.
			$Shell: {
				command: 'git tag -a v${Package.version} -m "Version v${Package.version}"',
				out: { console: true },
				err: { console: true },
				halt_on_error: false
			}
		},
		{ // Push the tag to Github.
			$Shell: {
				command: 'git push --quiet origin v${Package.version}',
				out: { console: true },
				err: { console: true },
				halt_on_error: false
			}
		},
	],
};
```


### Running the Task

```shell
npx @liquicode/devops --tasks my-tasks.js git_publish_version
```


### Discussion

The task file contains a `Context` section which allows you to initialize the task context with specific values.
In this case, we are loading the contents of the `package.json` file in to the context field called `Package`.
We can then refer to package data in our tasks like `${Package.version}` and `devops` will replace the values for us.

This task file defines a single task called `git_publish_version` and contains an array of steps/commands.
Steps are executed in the order that they appear in the task.
If one step fails or generates an error, then `devops` will stop processing the task.
This task issues several `git` commands to push all of our code changes to Github and then create a version tag to mark it.
When `halt_on_error: false` is set for a `$Shell` command, any errors generated by the shell command will not halt the processing of this task, as it otherwise would.
These steps also redirect the shell's `stdout` and `stderr` to the console.

In order to execute this task, you give `devops` the task file and the name of the task to run.
Using the npm utility `npx` will run `devops` whether it is installed locally in your project or globally in your environment.
If not installed at all, `npx` will temporarily download and run the latest version of `devops`.

